'----------------------------------------------------------------------------------
' Program:       SRM9000 Radio Dispatcher
' Module:        FormatPmr
' Description:   This module
'                  - Builds the appropriate binary serial string for   
'                    sending to the radio.
'                  - Extracts the message parameters from a received
'                    binary serial message from the radio.
'
'
'
' This software is provided as Freeware by TMC Radio Pty Ltd.
' This software and information is provided 'As Is' without warranty of any
' kind, either expressed or implied. 
'
'----------------------------------------------------------------------------------

Imports System.Text


Module FormatPmr

    'PMR Radio Rx Commands
    Const CMD_STATE_REPORT As Integer = &H9B         '155 0x9B
    Const CMD_DATA_SEND_ACK As Integer = &HB2        '178 0xB2 
    Const CMD_SEND_PMR_DATA As Integer = &HB3        '179 0xB3
    Const CMD_RX_PMR_DATA As Integer = &HB4          '180 0xB4
    Const CMD_CONSOLE_TYPE As Integer = &HB5         '181 0xB5 
    Const CMD_REPORT_STATE As Integer = &HB9         '185 0xB9
    Const CMD_ERROR_REPORT As Integer = &HC7         '199 0xC7 

    'PMR Radio Tx Commands
    Const CMD_SEND_DATA_MSG As Byte = 179            '179 0xB3

    Public Function SerialTxPmrFormat(ByVal MsgParams As TXMESSAGEPARAMETERS) As Byte()

        'Convert structure of Message parameters into a binary serial string,
        'which can be sent to the dispatcher radio.
        'Send Data Message. Command 179 (0xB3)
        ' MsgLength, 0, Command, Destination[2], MsgFormat, Data[237]

        Dim MessagePkt(256) As Byte
        Dim DestinationMSB As Integer
        Dim DestinationLSB As Integer
        Dim MsgLength As Integer
        Dim ascii As New ASCIIEncoding
        Dim TempDest As Integer

        'Header
        MessagePkt(0) = 0    'Message Length, work out later
        MessagePkt(1) = 0
        MessagePkt(2) = CMD_SEND_DATA_MSG

        TempDest = CInt(MsgParams.Destination)
        DestinationMSB = TempDest >> 8
        MessagePkt(3) = CByte(DestinationMSB)
        DestinationLSB = CInt(CInt(MsgParams.Destination) Mod CInt(256))
        MessagePkt(4) = CByte(DestinationLSB)

        'Message Format
        'The serial comms document specifies around 16 message formats
        'This application only supports 3

        If MsgParams.MessageFormat = "STA" Then
            'Data is Status number only
            MessagePkt(5) = 10
            MessagePkt(6) = CByte(MsgParams.Status)
            'Total Length of MessagePkt is 6 bytes
            MessagePkt(0) = 6

        ElseIf MsgParams.MessageFormat = "DM" Then
            'ASCII data with Bip Bip Alert
            MessagePkt(5) = 2
            MsgLength = MsgParams.Message.Length
            ascii.GetBytes(MsgParams.Message, 0, MsgLength, MessagePkt, 6)
            'Total Length of MessagePkt is (MsgLength + 5) bytes
            MessagePkt(0) = CByte(MsgLength + 5)

        ElseIf MsgParams.MessageFormat = "PR" Then
            'GPS Poll Request
            MessagePkt(5) = 8
            MessagePkt(6) = &HFF
            MessagePkt(7) = &HFF
            'Total Length of MessagePkt is 7 bytes
            MessagePkt(0) = 7

        Else
            'Unkown type, so default send message as plain ascii data
            'ASCII data with Bip Bip Alert
            MessagePkt(5) = 2
            MsgLength = MsgParams.Message.Length
            ascii.GetBytes(MsgParams.Message, 0, MsgLength, MessagePkt, 6)
            MessagePkt(0) = CByte(MsgLength + 5)

        End If

        'Return Formated Binary serial message
        SerialTxPmrFormat = MessagePkt

    End Function

    Public Function SerialRxPmrFormat(ByVal MessagePkt As Byte()) As TXMESSAGEPARAMETERS

        'Convert received binary serial string to a structure of Message parameters,
        'which can later be displayed in the program's dispatcher list.
        'Received Data Message. Command 180 (0xB4)
        ' MsgLength|0|Command|Destination[2]|MsgFormat| Data[237] |

        Dim MsgParams As TXMESSAGEPARAMETERS
        Dim MessageType As Integer
        Dim MessageLength As Integer = MessagePkt(0)
        Dim DestinationMSB As Integer
        Dim DestinationLSB As Integer
        Dim Destination As Integer
        Dim ascii As New ASCIIEncoding

        Dim GPSUnpacked(30) As Byte
     
        MsgParams = Nothing


        Select Case MessagePkt(1)
            Case 0
                'PMR Message
                Select Case MessagePkt(2)

                    Case CMD_DATA_SEND_ACK
                        'Received acknowledge to sent message
                        MsgParams.Destination = ""
                        MsgParams.MessageFormat = "ACK"
                        MsgParams.AlertTone = ""
                        MsgParams.JobNumber = ""
                        MsgParams.Status = ""
                        MsgParams.Message = "Message Delivered"

                    Case CMD_RX_PMR_DATA
                        'Received data message, now determine the format of 
                        'the data message and create reply structure.
                        'Note: Destination is actually message source.
                        DestinationMSB = CInt(MessagePkt(3))
                        DestinationLSB = CInt(MessagePkt(4))
                        Destination = DestinationMSB * 256 + DestinationLSB
                        MsgParams.Destination = CStr(Destination)

                        MessageType = CInt(MessagePkt(5))
                        If MessageType <= 7 Then
                            'ASCII Message
                            MsgParams.MessageFormat = "Data " + CStr(MessageType)
                            MsgParams.AlertTone = ""
                            MsgParams.JobNumber = ""
                            MsgParams.Status = ""
                            MsgParams.Message = ascii.GetString(MessagePkt, 6, (MessageLength - 5))

                        ElseIf MessageType = 10 Then
                            'Status
                            MsgParams.MessageFormat = "STA" + CStr(MessageType)
                            MsgParams.AlertTone = ""
                            MsgParams.JobNumber = ""
                            MsgParams.Status = CStr(MessagePkt(6))
                            MsgParams.Message = "Status Message Received"

                        ElseIf MessageType = 11 Then
                            'Status + GPS
                            ' MsgParams.MessageFormat = "STA" + CStr(MessageType)
                            '  MsgParams.AlertTone = ""
                            '  MsgParams.JobNumber = ""
                            '  MsgParams.Status = CStr(MessagePkt(6))
                            'Convert NMEA-0183-GGA GPS Format to TMC Radio AVL database structure
                            ' MsgParams.Message = ConvertGgaToAvl(MessagePkt, SetupOptions.RadioMode, 11, 0)

                        ElseIf MessageType = 12 Then
                            'Status + Job Number
                            MsgParams.MessageFormat = "Data " + CStr(MessageType)
                            MsgParams.AlertTone = ""
                            MsgParams.JobNumber = ascii.GetString(MessagePkt, 7, 8)
                            MsgParams.Status = CStr(MessagePkt(6))
                            MsgParams.Message = "Status + Job Number"

                        ElseIf MessageType = 13 Then
                            'Status + Job Number + GPS
                            ' MsgParams.MessageFormat = "Data " + CStr(MessageType)
                            ' MsgParams.AlertTone = ""
                            '' MsgParams.JobNumber = ascii.GetString(MessagePkt, 7, 8)
                            ' MsgParams.Status = CStr(MessagePkt(6))
                            'Convert NMEA-0183-GGA GPS Format to TMC Radio AVL database structure
                            ' MsgParams.Message = ConvertGgaToAvl(MessagePkt, SetupOptions.RadioMode, 13, 0)

                        ElseIf MessageType = 14 Then
                            'Status + GPS 
                            'MsgParams.MessageFormat = "Data " + CStr(MessageType)
                            ' MsgParams.AlertTone = ""
                            ' MsgParams.JobNumber = ""
                            'MsgParams.Status = CStr(MessagePkt(6))
                            'Convert NMEA-0183-RMC GPS Format to TMC Radio AVL database structure
                            'MsgParams.Message = ConvertRmcToAvl(MessagePkt, SetupOptions.RadioMode, 14, 0)

                        ElseIf MessageType = 15 Then
                            'Status + GPS + Job Number
                            '  MsgParams.MessageFormat = "Data " + CStr(MessageType)
                            ' MsgParams.AlertTone = ""
                            ' MsgParams.JobNumber = ascii.GetString(MessagePkt, 20, 7)
                            ' MsgParams.Status = CStr(MessagePkt(6))
                            ' 'Convert NMEA-0183-RMC GPS Format to TMC Radio AVL database structure
                            ' MsgParams.Message = ConvertRmcToAvl(MessagePkt, SetupOptions.RadioMode, 15, 0)
                            '
                        Else
                            'ASCII Message
                            MsgParams.MessageFormat = "Data " + CStr(MessageType)
                            MsgParams.AlertTone = ""
                            MsgParams.JobNumber = ""
                            MsgParams.Status = ""
                            MsgParams.Message = "Unknown Data Message Format"
                        End If

                    Case CMD_ERROR_REPORT
                        'Received Error from last sent message
                        MsgParams.Destination = ""
                        MsgParams.MessageFormat = "Error"
                        MsgParams.AlertTone = ""
                        MsgParams.JobNumber = ""
                        MsgParams.Status = ""
                        Select Case MessagePkt(3)
                            Case 0
                                MsgParams.Message = "Invaild Message: Unknown Command"
                            Case 1
                                MsgParams.Message = "Invalid Message: Bad Length"
                            Case 2
                                MsgParams.Message = "Transmit not allowed on current radio channel"
                            Case 3 To 9
                                MsgParams.Message = "Cannot Transmit"
                            Case 10
                                MsgParams.Message = "No response from called radio"
                        End Select
                End Select

            Case 8
                'MAP27 Messages
                Select Case MessagePkt(2)
                    Case 0
                        'Link Established
                    Case 1
                        'Link Lost
                End Select

        End Select

        'Return message structure
        SerialRxPmrFormat = MsgParams

    End Function

End Module
