'----------------------------------------------------------------------------------
' Program:       SRM9000 Radio Dispatcher
' Module:        FormatTrunk
' Description:   This module
'                  - Builds the appropriate binary serial string for   
'                    sending to the radio.
'                  - Extracts the message parameters from a received
'                    binary serial message from the radio.
'
'
'
' This software is provided as Freeware by TMC Radio Pty Ltd.
' This software and information is provided 'As Is' without warranty of any
' kind, either expressed or implied. 
'
'----------------------------------------------------------------------------------

Imports System.Text

Module FormatTrunk
    'MAP27 Trunk Radio Rx Commands
    Const CMD_RX_STATUS As Integer = &H80
    Const CMD_ACK_POSITIVE As Integer = &HC0
    Const CMD_ACK_QUEUING As Integer = &HD0
    Const CMD_ACK_NEGATIVE As Integer = &HE0
    Const CMD_RECEIVE_SST As Integer = &H81
    Const CMD_RECEIVE_MST As Integer = &H82
    Const CMD_NUMBERING_INFO As Integer = &HB1
    Const CMD_OPERATING_CONDITION As Integer = &HB2
    Const CMD_SETUP_PROGRESS_POS As Integer = &HC4
    Const CMD_SETUP_PROGRESS_QUE As Integer = &HD4
    Const CMD_SETUP_PROGRESS_NEG As Integer = &HE4
    Const CMD_CALL_CLEARED As Integer = &HA6

    'MAP27 Trunk Radio Tx Commands
    Const CMD_SEND_STATUS As Byte = &H80
    Const CMD_SEND_DATA As Byte = &H82
    Const CMD_START_NPD_CALL As Integer = &HA4
    Const CMD_END_NPD_CALL As Integer = &H86    'Disconnect Call
    Const CMD_MODEM_DATA As Byte = &HA3         'Not used

    'SRM9000 Serial Commands
    Const CMD_SRM9000_SERIAL As Byte = &H0      'SRM9000 serial command 
    Const CMD_SEND_DATA_MSG As Byte = &HB3      'SRM9000 data message
    Const CMD_DATA_SEND_ACK As Byte = &HB2      'Data send acknowledge  
    Const CMD_ERROR_REPORT As Byte = &HC7       'Error Report  


    Public Function SerialTxTrunkFormat(ByVal MsgParams As TXMESSAGEPARAMETERS) As Byte()

        'Convert structure of Message parameters into a binary serial string,
        'which can be sent to the dispatcher radio.

        Dim MessagePkt(256) As Byte
        Dim MsgLength As Integer
        Dim ascii As New ASCIIEncoding
        Dim Prefix, Ident, IdentA, IdentB As Integer
        Dim CalledUnitNumber As Integer


        'We know the destination radio's unit number, and we know the dispatcher 
        'radio's IBI and Fleet size, so we can now work out Destination radio's 
        'MPT1327 Identity (Prefix, IdentA and IdentB).
        'It is assumed that the caller is in the same fleet as the dispatcher radio.

        If OwnFleetSize = LARGE_FLEET Then
            'Large Fleet Individual or Group Identity
            Prefix = OwnPrefix
            CalledUnitNumber = CInt(MsgParams.Destination)
            If (CalledUnitNumber >= 200) And (CalledUnitNumber <= 899) Then
                Ident = CalledUnitNumber + OwnIBI - 200
            ElseIf (CalledUnitNumber >= 900) And (CalledUnitNumber <= 998) Then
                Ident = CalledUnitNumber + OwnGBI - 900
            Else
                'Invalid Ident
                Ident = 0
            End If
            IdentA = CInt(Ident >> 8)
            IdentB = CInt(Ident Mod 256)
        Else
            'Small Fleet Individual or Group Identity 
            Prefix = OwnPrefix
            CalledUnitNumber = CInt(MsgParams.Destination)
            If (CalledUnitNumber >= 20) And (CalledUnitNumber <= 89) Then
                Ident = CalledUnitNumber + OwnIBI - 20
            ElseIf (CalledUnitNumber >= 90) And (CalledUnitNumber <= 99) Then
                Ident = CalledUnitNumber + OwnGBI - 90
            Else
                'Invalid Ident
                Ident = 0
            End If
            IdentA = CInt(Ident >> 8)
            IdentB = CInt(Ident Mod 256)
        End If

        'Build Data Message
        If MsgParams.MessageFormat = "STA" Then
            MessagePkt(0) = CByte(6)                'Message Length  
            MessagePkt(1) = CMD_SEND_STATUS         'Command = Send Status Message
            MessagePkt(2) = CByte(Prefix)
            MessagePkt(3) = CByte(IdentA)
            MessagePkt(4) = CByte(IdentB)
            MessagePkt(5) = 0                       'ADESC and Length
            MessagePkt(6) = CByte(MsgParams.Status)

        ElseIf MsgParams.MessageFormat = "DM" Then
            MsgLength = MsgParams.Message.Length
            If MsgLength > 100 Then
                'Maximum Length of EDM message is 100 Characters
                MsgLength = 100
            End If
            MessagePkt(0) = CByte(MsgLength + 6)
            MessagePkt(1) = CMD_SEND_DATA           'Command = Send Data Message 
            MessagePkt(2) = CByte(Prefix)
            MessagePkt(3) = CByte(IdentA)
            MessagePkt(4) = CByte(IdentB)
            MessagePkt(5) = 0                       'ADESC and Length
            MessagePkt(6) = &HB7                    'Coding and NLB
            ascii.GetBytes(MsgParams.Message, 0, MsgLength, MessagePkt, 7)

        ElseIf MsgParams.MessageFormat = "PR" Then
            MessagePkt(0) = CByte(6)                'Message Length  
            MessagePkt(1) = CMD_SEND_STATUS         'Command = Send Poll Request (Status Message)
            MessagePkt(2) = CByte(Prefix)
            MessagePkt(3) = CByte(IdentA)
            MessagePkt(4) = CByte(IdentB)
            MessagePkt(5) = 0                       'ADESC and Length
            MessagePkt(6) = CByte(SetupOptions.PollRequestStatus + 1) 'Poll Request Status Number

        ElseIf MsgParams.MessageFormat = "NPD_On" Then
            'Start an NPD call so Modem Data can be sent on a traffic channel
            MessagePkt(0) = CByte(6)                'Message Length  
            MessagePkt(1) = CMD_START_NPD_CALL      'Command = Initiate an NPD Call
            MessagePkt(2) = CByte(Prefix)
            MessagePkt(3) = CByte(IdentA)
            MessagePkt(4) = CByte(IdentB)
            MessagePkt(5) = 0                       'ADESC and Length
            MessagePkt(6) = &H18                    'Modem Call 

        ElseIf MsgParams.MessageFormat = "MD" Then
            'Once a Non-Prescribed-Data (NPD) call has been setup (ie trunk traffic 
            'channel allocated between caller and called radio), modem data of any type  
            'and length (237 bytes at a time) can be sent on the traffic channel.
            'For this dispatcher we are going to send data in a format that a receiving
            'SRM9000 or SRP9100 can display.
            MsgLength = MsgParams.Message.Length
            MessagePkt(0) = CByte(MsgLength + 5)    'Message Length 
            MessagePkt(1) = 0
            MessagePkt(2) = CMD_SEND_DATA_MSG
            MessagePkt(3) = CByte(IdentA)
            MessagePkt(4) = CByte(IdentB)
            MessagePkt(5) = 2                       'ASCII data 
            ascii.GetBytes(MsgParams.Message, 0, MsgLength, MessagePkt, 6)

        ElseIf MsgParams.MessageFormat = "NPD_Off" Then
            'End NPD Call
            MessagePkt(0) = CByte(6)                'Message Length  
            MessagePkt(1) = CMD_END_NPD_CALL        'Command = Disconnect Call
            MessagePkt(2) = CByte(Prefix)
            MessagePkt(3) = CByte(IdentA)
            MessagePkt(4) = CByte(IdentB)
            MessagePkt(5) = 0                       'ADESC and Length
            MessagePkt(6) = &H7                     'Normal End

        End If

        'Return Formated Binary serial message
        SerialTxTrunkFormat = MessagePkt

    End Function

    Public Function SerialRxTrunkFormat(ByVal MessagePkt As Byte()) As TXMESSAGEPARAMETERS

        'Convert received binary serial string to a structure of Message parameters,
        'which can later be displayed in the program's dispatcher list.

        Dim MsgParams As TXMESSAGEPARAMETERS
        Dim MessageType As Integer
        Dim MessageLength As Integer = MessagePkt(0)

        Dim Prefix, IdentA, IdentB As Integer
        Dim CallersUnitNumber As Integer
        Dim AdescLength As Integer
        Dim Status As Integer
        Dim Reason As String
        Dim DataLength, DataType, DataOffset As Integer

        Dim ascii As New ASCIIEncoding
        Dim FleetSize As String

        Dim GPSUnpacked(30) As Byte
        Reason = ""
        MsgParams = Nothing


        Select Case MessagePkt(1)
            'Trunk CommandMessage 

            Case CMD_RX_STATUS
                'Received Trunk Status Message
                '|Length|Command|Prefix|IdentA|IdentB|Adesc|..|Status|
                Prefix = CInt(MessagePkt(2))
                IdentA = CInt(MessagePkt(3))
                IdentB = CInt(MessagePkt(4))
                'Get caller's Unit number from MPT1327 Prefix\Ident
                'It is assumed that the caller is in the same fleet as the dispatcher radio
                If OwnFleetSize = LARGE_FLEET Then
                    CallersUnitNumber = (IdentA * 256) + IdentB - OwnIBI + 200
                Else
                    CallersUnitNumber = (IdentA * 256) + IdentB - OwnIBI + 20
                End If

                AdescLength = CInt(MessagePkt(5))
                Status = CInt(MessagePkt(6) + AdescLength)

                'Create Function Reply Structure
                MsgParams.Destination = CStr(CallersUnitNumber)
                MsgParams.MessageFormat = "STA"
                MsgParams.AlertTone = ""
                MsgParams.JobNumber = ""
                MsgParams.Status = CStr(Status)
                MsgParams.Message = "Status Message Received"

            Case CMD_RECEIVE_SST, CMD_RECEIVE_MST
                'Received Data Message
                '|Length|Command|Prefix|IdentA|IdentB|Adesc|..|NLB|Data..|
                Prefix = CInt(MessagePkt(2))
                IdentA = CInt(MessagePkt(3))
                IdentB = CInt(MessagePkt(4))
                'Get caller's Unit number from MPT1327 Prefix\Ident
                'It is assumed that the caller is in the same fleet as the dispatcher radio
                If OwnFleetSize = LARGE_FLEET Then
                    CallersUnitNumber = (IdentA * 256) + IdentB - OwnIBI + 200
                Else
                    CallersUnitNumber = (IdentA * 256) + IdentB - OwnIBI + 20
                End If

                AdescLength = CInt(MessagePkt(5)) Mod 16                'Low Nibble
                DataLength = (MessagePkt(5 + AdescLength + 1)) Mod 16   'Low Nibble 

                DataType = CInt((MessagePkt(5 + AdescLength + 1)) >> 4) 'High Nibble 
                DataOffset = 5 + AdescLength + 2
                DataLength = 0
                Do While CBool(MessagePkt(DataOffset + DataLength))
                    DataLength += 1
                Loop

                MsgParams.MessageFormat = "Data"
                If DataType = &HB Then
                    'ASCII Data Message
                    MsgParams.Message = ascii.GetString(MessagePkt, DataOffset, DataLength)

                ElseIf DataType = &H8 Then
                    'Binary Data Message: Special SST Format
                    ' |8n|Type|Status|JobNo[8]|GPSData[12]|

                    MessageType = MessagePkt(DataOffset)
                    If MessagePkt(DataOffset) = 1 Then
                        '  'Status + GPS
                        '  MsgParams.MessageFormat = "Data " + CStr(MessageType)
                        '  MsgParams.AlertTone = ""
                        '  MsgParams.JobNumber = ""
                        '  MsgParams.Status = CStr(MessagePkt(DataOffset + 1))
                        'Convert NMEA-0183-GGA GPS Format to TMC Radio AVL database structure
                        '  MsgParams.Message = ConvertGgaToAvl(MessagePkt, SetupOptions.RadioMode, 1, DataOffset)

                    ElseIf MessagePkt(DataOffset) = 2 Then
                        'Status + Job Number
                        MsgParams.MessageFormat = "Data " + CStr(MessageType)
                        MsgParams.AlertTone = ""
                        MsgParams.JobNumber = ascii.GetString(MessagePkt, DataOffset + 2, 8)
                        MsgParams.Status = CStr(MessagePkt(DataOffset + 1))
                        MsgParams.Message = "Status + Job Number"

                    ElseIf MessagePkt(DataOffset) = 3 Then
                        'Status + Job Number + GPS
                        '  MsgParams.MessageFormat = "Data " + CStr(MessageType)
                        '  MsgParams.AlertTone = ""
                        '  MsgParams.JobNumber = ascii.GetString(MessagePkt, DataOffset + 2, 8)
                        '  MsgParams.Status = CStr(MessagePkt(DataOffset + 1))
                        'Convert NMEA-0183-GGA GPS Format to TMC Radio AVL database structure
                        '  MsgParams.Message = ConvertGgaToAvl(MessagePkt, SetupOptions.RadioMode, 3, DataOffset)

                    ElseIf MessagePkt(DataOffset) = 14 Then
                        'Status + GPS
                        '  MsgParams.MessageFormat = "Data " + CStr(MessageType)
                        '  MsgParams.AlertTone = ""
                        '  MsgParams.JobNumber = ""
                        '  MsgParams.Status = ""
                        'Convert NMEA-0183-RMC GPS Format to TMC Radio AVL database structure
                        '   MsgParams.Message = ConvertRmcToAvl(MessagePkt, SetupOptions.RadioMode, 14, DataOffset)

                    ElseIf MessagePkt(DataOffset) = 15 Then
                        'Status + GPS + Job Number 
                        ' MsgParams.MessageFormat = "Data " + CStr(MessageType)
                        '  MsgParams.AlertTone = ""
                        '  MsgParams.JobNumber = ""
                        '  MsgParams.Status = ""
                        'Convert NMEA-0183-RMC GPS Format to TMC Radio AVL database structure
                        '   MsgParams.Message = ConvertRmcToAvl(MessagePkt, SetupOptions.RadioMode, 15, DataOffset)

                    End If

                Else
                    'Binary Data Message: Unformated
                    MsgParams.Message = ascii.GetString(MessagePkt, DataOffset, DataLength)
                End If

                'Create Function Reply Structure
                MsgParams.Destination = CStr(CallersUnitNumber)
                'MsgParams.MessageFormat = "Data"
                MsgParams.AlertTone = ""
                'MsgParams.JobNumber = ""
                'MsgParams.Status = ""
                'MsgParams.Message = ""

            Case CMD_ACK_POSITIVE, CMD_SETUP_PROGRESS_POS
                'Call Connected Acknowledge
                '|Length|Command|Prefix|IdentA|IdentB|Adesc|..|Reason|

                'Create Function Reply Structure
                MsgParams.Destination = ""
                MsgParams.MessageFormat = "ACKP"
                MsgParams.AlertTone = ""
                MsgParams.JobNumber = ""
                MsgParams.Status = ""
                MsgParams.Message = "Message Delivered"

            Case CMD_ACK_QUEUING, CMD_SETUP_PROGRESS_QUE
                'Message Queued
                '|Length|Command|Prefix|IdentA|IdentB|Adesc|..|Reason|

                Prefix = CInt(MessagePkt(2))
                IdentA = CInt(MessagePkt(3))
                IdentB = CInt(MessagePkt(4))
                'Get caller's Unit number from MPT1327 Prefix\Ident
                'It is assumed that the caller is in the same fleet as the dispatcher radio
                If OwnFleetSize = LARGE_FLEET Then
                    CallersUnitNumber = (IdentA * 256) + IdentB - OwnIBI + 200
                Else
                    CallersUnitNumber = (IdentA * 256) + IdentB - OwnIBI + 20
                End If

                AdescLength = CInt(MessagePkt(5))
                'Find out why call is being queued
                Select Case MessagePkt(6 + AdescLength)
                    Case 1
                        Reason = "Called Unit Ringing"
                    Case 2
                        Reason = "Call Queued - System Busy"
                    Case &HA
                        Reason = "Called Unit Engaged"
                    Case &H26
                        Reason = "Called Unit Diverted"
                    Case &H22
                        Reason = "TSC doesnt support Data"
                    Case Else
                        Reason = "Call Queued - Reason Unknown"
                End Select

                'Create Function Reply Structure
                MsgParams.Destination = CStr(CallersUnitNumber)
                MsgParams.MessageFormat = "TNM_QUE"    'Trunk Network Message
                MsgParams.AlertTone = ""
                MsgParams.JobNumber = ""
                MsgParams.Status = ""
                MsgParams.Message = Reason

            Case CMD_ACK_NEGATIVE, CMD_SETUP_PROGRESS_NEG
                'Message or command not Acknowledged
                '|Length|Command|Prefix|IdentA|IdentB|Adesc|..|Reason|

                Prefix = CInt(MessagePkt(2))
                IdentA = CInt(MessagePkt(3))
                IdentB = CInt(MessagePkt(4))
                'Get caller's Unit number from MPT1327 Prefix\Ident
                'It is assumed that the caller is in the same fleet as the dispatcher radio
                If OwnFleetSize = LARGE_FLEET Then
                    CallersUnitNumber = (IdentA * 256) + IdentB - OwnIBI + 200
                Else
                    CallersUnitNumber = (IdentA * 256) + IdentB - OwnIBI + 20
                End If

                AdescLength = CInt(MessagePkt(5))
                'Find out why call failed
                Select Case MessagePkt(6 + AdescLength)
                    Case 3
                        Reason = "Call Failed - Invalid Call"
                    Case &H4
                        Reason = "Called Radio Out Of Reach"
                    Case &H6
                        Reason = "Called Unit Diverted"
                    Case 7
                        Reason = "Emergency Call - Please Wait"
                    Case &H8
                        Reason = "Transaction Aborted"
                    Case &HB
                        Reason = "Call Failed - System Overload"
                    Case &HC
                        Reason = "Called Unit Engaged"
                    Case &HE
                        Reason = "Called Radio Diverted to Unknown Address"
                    Case &H16
                        Reason = "Called Radio Diverted to Unknown Group"
                    Case Else
                        Reason = "Call Failed"
                End Select

                'Create Function Reply Structure
                MsgParams.Destination = CStr(CallersUnitNumber)
                MsgParams.MessageFormat = "ACKN"
                MsgParams.AlertTone = ""
                MsgParams.JobNumber = ""
                MsgParams.Status = ""
                MsgParams.Message = Reason

            Case CMD_CALL_CLEARED, CMD_END_NPD_CALL
                'Call Connected Acknowledge
                '|Length|Command|Prefix|IdentA|IdentB|Reason|

                'Create Function Reply Structure
                MsgParams.Destination = ""
                MsgParams.MessageFormat = "ACKP"
                MsgParams.AlertTone = ""
                MsgParams.JobNumber = ""
                MsgParams.Status = ""
                MsgParams.Message = ""

            Case CMD_NUMBERING_INFO
                'Dispatcher Radio's Numbering Information
                '|Length|Command|Prefix|Ident[2]|IBI[2]|FleetSize|HighestIdent[2]|
                '  |GrpPrefix|GrpIdent[2]|GBI[2]|GrpFleetSize|GrpHighestIdent[2]|

                OwnPrefix = CInt(MessagePkt(2))
                OwnIdent = CInt(MessagePkt(3) * 256) + MessagePkt(4)
                OwnIBI = CInt((MessagePkt(5) Mod 128) * 256) + MessagePkt(6)
                OwnFleetSize = CInt(MessagePkt(5) >> 7)
                OwnHighestIdent = CInt(MessagePkt(7) * 256) + MessagePkt(8)

                OwnGrpPrefix = CInt(MessagePkt(13))
                OwnGrpIdent = CInt(MessagePkt(14) * 256) + MessagePkt(15)
                OwnGBI = CInt((MessagePkt(8) Mod 128) * 256) + MessagePkt(9)
                OwnGrpFleetSize = CInt(MessagePkt(8) >> 7)
                OwnGrpHighestIdent = CInt(MessagePkt(10) * 256) + MessagePkt(11)

                If OwnFleetSize > 0 Then
                    FleetSize = "Large"
                Else
                    FleetSize = "Small"
                End If

                'Create Function Reply Structure
                MsgParams.Destination = CStr(CallersUnitNumber)
                MsgParams.MessageFormat = "TNM"    'Trunk Network Message
                MsgParams.AlertTone = ""
                MsgParams.JobNumber = ""
                MsgParams.Status = ""
                MsgParams.Message = "Dispatcher: " + CStr(OwnPrefix) + "/" + CStr(OwnIdent) + _
                                    "  IBI: " + CStr(OwnIBI) + "  FleetSize: " + FleetSize

            Case CMD_OPERATING_CONDITION
                ''Only interested in the Flag Byte.
                '|Length|Command|Flags|RSSI|MCDT|
                'Flag Byte = |  |  |  |  |  |Service|Busy|Transmit|

                Dim TempFlagBit As Byte

                TempFlagBit = CByte(MessagePkt(2) And &H4)
                If TempFlagBit = 4 Then
                    TrunkServiceIndicator = IN_SERVICE
                Else
                    TrunkServiceIndicator = OUT_OF_SERVICE
                End If

                TempFlagBit = CByte(MessagePkt(2) And &H2)
                If TempFlagBit = 2 Then
                    TrunkBusyIndicator = BUSY
                Else
                    TrunkBusyIndicator = NOT_BUSY
                End If

                TempFlagBit = CByte(MessagePkt(1) And &H1)
                If TempFlagBit = 1 Then
                    TrunkTransmitIndicator = TRANSMIT
                Else
                    TrunkTransmitIndicator = RECEIVE
                End If

                'Create Function Reply Structure
                MsgParams.Destination = ""
                MsgParams.MessageFormat = "TNM_OP"    'Trunk Network Message
                MsgParams.AlertTone = ""
                MsgParams.JobNumber = ""
                MsgParams.Status = ""
                MsgParams.Message = ""

            Case CMD_SRM9000_SERIAL
                'SRM9000 serial commands are defined by the SUB-COMMAND in 
                'the following byte.
                If MessagePkt(2) = CMD_DATA_SEND_ACK Then
                    'NPD data message has successfully been sent
                    Reason = "Message Delivered"

                ElseIf MessagePkt(2) = CMD_ERROR_REPORT Then
                    'NPD data message was not correctly received by receiving radio.
                    Select Case MessagePkt(3)
                        Case 0
                            Reason = "Error: Unknown command byte"
                        Case 1
                            Reason = "Error: Unable to execute. Invalid parameter or radio busy"
                        Case 2
                            Reason = "Error: Bad length"
                        Case 3 To 9
                            Reason = "Error: Cannot Transmit"
                        Case 10
                            Reason = "No acknowledge from called radio"
                    End Select

                End If

                'Create Function Reply Structure
                MsgParams.Destination = CStr(CallersUnitNumber)
                MsgParams.MessageFormat = "NPD"
                MsgParams.AlertTone = ""
                MsgParams.JobNumber = ""
                MsgParams.Status = ""
                MsgParams.Message = Reason

        End Select

        'Return Function Reply Structure
        SerialRxTrunkFormat = MsgParams

    End Function


End Module
